import os
import torch
import warnings
import torchaudio
from typing import List
from itertools import groupby


def read_batch(audio_paths: List[str]):
    return [read_audio(audio_path)
            for audio_path
            in audio_paths]


def split_into_batches(lst: List[str],
                       batch_size: int = 10):
    return [lst[i:i + batch_size]
            for i in
            range(0, len(lst), batch_size)]


def read_audio(path: str,
               target_sr: int = 16000):

    wav, sr = torchaudio.load(path)

    if wav.size(0) > 1:
        wav = wav.mean(dim=0, keepdim=True)

    if sr != target_sr:
        transform = torchaudio.transforms.Resample(orig_freq=sr,
                                                   new_freq=target_sr)
        wav = transform(wav)
        sr = target_sr

    assert sr == target_sr
    return wav.squeeze(0)


def prepare_model_input(batch: List[torch.Tensor],
                        device=torch.device('cpu')):
    max_seqlength = max(max([len(_) for _ in batch]), 12800)
    inputs = torch.zeros(len(batch), max_seqlength)
    for i, wav in enumerate(batch):
        inputs[i, :len(wav)].copy_(wav)
    inputs = inputs.to(device)
    return inputs


class Decoder():
    def __init__(self,
                 labels: List[str]):
        self.labels = labels
        self.blank_idx = self.labels.index('_')
        self.space_idx = self.labels.index(' ')

    def process(self,
                probs, wav_len, word_align):
        assert len(self.labels) == probs.shape[1]
        for_string = []
        argm = torch.argmax(probs, axis=1)
        align_list = [[]]
        for j, i in enumerate(argm):
            if i == self.labels.index('2'):
                try:
                    prev = for_string[-1]
                    for_string.append('$')
                    for_string.append(prev)
                    align_list[-1].append(j)
                    continue
                except:
                    for_string.append(' ')
                    warnings.warn('Token "2" detected a the beginning of sentence, omitting')
                    align_list.append([])
                    continue
            if i != self.blank_idx:
                for_string.append(self.labels[i])
                if i == self.space_idx:
                    align_list.append([])
                else:
                    align_list[-1].append(j)

        string = ''.join([x[0] for x in groupby(for_string)]).replace('$', '').strip()

        align_list = list(filter(lambda x: x, align_list))

        if align_list and wav_len and word_align:
            align_dicts = []
            linear_align_coeff = wav_len / len(argm)
            to_move = min(align_list[0][0], 1.5)
            for i, align_word in enumerate(align_list):
                if len(align_word) == 1:
                    align_word.append(align_word[0])
                align_word[0] = align_word[0] - to_move
                if i == (len(align_list) - 1):
                    to_move = min(1.5, len(argm) - i)
                    align_word[-1] = align_word[-1] + to_move
                else:
                    to_move = min(1.5, (align_list[i+1][0] - align_word[-1]) / 2)
                    align_word[-1] = align_word[-1] + to_move

            for word, timing in zip(string.split(), align_list):
                align_dicts.append({'word': word,
                                    'start_ts': round(timing[0] * linear_align_coeff, 2),
                                    'end_ts': round(timing[-1] * linear_align_coeff, 2)})

            return string, align_dicts
        return string

    def __call__(self,
                 probs: torch.Tensor,
                 wav_len: float = 0,
                 word_align: bool = False):
        return self.process(probs, wav_len, word_align)


def init_jit_model(model_url: str,
                   device: torch.device = torch.device('cpu')):
    torch.set_grad_enabled(False)

    model_dir = os.path.join(os.path.dirname(__file__), "model")
    os.makedirs(model_dir, exist_ok=True)
    model_path = os.path.join(model_dir, os.path.basename(model_url))

    if not os.path.isfile(model_path):
        torch.hub.download_url_to_file(model_url,
                                       model_path,
                                       progress=True)

    model = torch.jit.load(model_path, map_location=device)
    model.eval()
    return model, Decoder(model.labels)



import os
import re
import torch
import warnings


def init_jit_model(model_url: str,
                   device: torch.device = torch.device('cpu')):
    torch.set_grad_enabled(False)

    model_dir = os.path.join(os.path.dirname(__file__), "model")
    os.makedirs(model_dir, exist_ok=True)
    model_path = os.path.join(model_dir, os.path.basename(model_url))

    if not os.path.isfile(model_path):
        torch.hub.download_url_to_file(model_url,
                                       model_path,
                                       progress=True)

    model = torch.jit.load(model_path, map_location=device)
    model.eval()
    return model


def prepare_text_input(text, symbols, symbol_to_id=None):
    if len(text) > 140:
        warnings.warn('Text string is longer than 140 symbols.')

    if symbol_to_id is None:
        symbol_to_id = {s: i for i, s in enumerate(symbols)}

    text = text.lower()
    text = re.sub(r'[^{}]'.format(symbols[2:]), '', text)
    text = re.sub(r'\s+', ' ', text).strip()
    if text[-1] not in ['.', '!', '?']:
        text = text + '.'
    text = text + symbols[1]

    text_ohe = [symbol_to_id[s] for s in text if s in symbols]
    text_tensor = torch.LongTensor(text_ohe)
    return text_tensor


def prepare_tts_model_input(text: str or list, symbols: str):
    if type(text) == str:
        text = [text]
    symbol_to_id = {s: i for i, s in enumerate(symbols)}
    if len(text) == 1:
        return prepare_text_input(text[0], symbols, symbol_to_id).unsqueeze(0), torch.LongTensor([0])

    text_tensors = []
    for string in text:
        string_tensor = prepare_text_input(string, symbols, symbol_to_id)
        text_tensors.append(string_tensor)
    input_lengths, ids_sorted_decreasing = torch.sort(
            torch.LongTensor([len(t) for t in text_tensors]),
            dim=0, descending=True)
    max_input_len = input_lengths[0]
    batch_size = len(text_tensors)

    text_padded = torch.ones(batch_size, max_input_len, dtype=torch.int32)

    for i, idx in enumerate(ids_sorted_decreasing):
        text_tensor = text_tensors[idx]
        in_len = text_tensor.size(0)
        text_padded[i, :in_len] = text_tensor

    return text_padded, ids_sorted_decreasing


def process_tts_model_output(out, out_lens, ids, sample_rate):
    assert sample_rate in [8000, 16000]
    out = out.to('cpu')
    out_lens = out_lens.to('cpu')
    _, orig_ids = ids.sort()

    proc_outs = []
    srf = 2 if sample_rate == 16000 else 1
    orig_out = out.index_select(0, orig_ids)
    orig_out_lens = out_lens.index_select(0, orig_ids)

    for i, out_len in enumerate(orig_out_lens):
        proc_outs.append(orig_out[i][:out_len*srf])
    return proc_outs


def apply_tts(texts: list,
              model: torch.nn.Module,
              sample_rate: int,
              symbols: str,
              device: torch.device):
    text_padded, orig_ids = prepare_tts_model_input(texts, symbols=symbols)
    out, out_lens = model(text_padded.to(device))
    audios = process_tts_model_output(out, out_lens, orig_ids, sample_rate)
    return audios



import os
import torch
import torchaudio


def read_audio(
    path: str,
    sampling_rate: int = 24000
):
    wav, sr = torchaudio.load(path)
    if wav.size(0) > 1:
        wav = wav.mean(dim=0, keepdim=True)
    if sr != sampling_rate:
        transform = torchaudio.transforms.Resample(
            orig_freq=sr,
            new_freq=sampling_rate
        )
        wav = transform(wav)
        sr = sampling_rate
    assert sr == sampling_rate
    return wav * 0.95


def save_audio(
    path: str,
    tensor: torch.Tensor,
    sampling_rate: int = 48000
):
    torchaudio.save(path, tensor, sampling_rate)


def init_jit_model(
    model_url: str,
    device: torch.device = torch.device('cpu')
):
    torch._C._jit_set_profiling_mode(False) 
    torch.set_grad_enabled(False)
    model_dir = os.path.join(os.path.dirname(__file__), "model")
    os.makedirs(model_dir, exist_ok=True)
    model_path = os.path.join(model_dir, os.path.basename(model_url))
    if not os.path.isfile(model_path):
        torch.hub.download_url_to_file(model_url,
                                       model_path,
                                       progress=True)
    model = torch.jit.load(model_path)
    model.eval()
    model = model.to(device)
    return model


def denoise(
    model: torch.nn.Module,
    audio_path: str,
    save_path: str = 'result.wav',
    device=torch.device('cpu'),
    sampling_rate: int = 48000
):
    audio = read_audio(audio_path).to(device)
    model.to(device)
    out = model(audio).flatten().unsqueeze(0)
    if save_path:
        save_audio(save_path, out.cpu(), sampling_rate)
    return out, sampling_rate



from .silero import (
    silero_stt,
    silero_tts,
    silero_te,
    silero_denoise,
)

__all__ = [
    "silero_stt",
    "silero_tts",
    "silero_te",
    "silero_denoise",
]

__version__ = "0.1.0"
__author__ = "Silero Team"



import os
import torch


def silero_stt(language='en',
               version='latest',
               jit_model='jit',
               **kwargs):
    """ Silero Speech-To-Text Model(s)
    language (str): language of the model, now available are ['en', 'de', 'es']
    Returns a model, decoder object and a set of utils
    Please see https://github.com/snakers4/silero-models for usage examples
    """
    from omegaconf import OmegaConf    
    from .utils import (init_jit_model,
                        read_audio,
                        read_batch,
                        split_into_batches,
                        prepare_model_input)

    models_list_file = os.path.join(os.path.dirname(__file__), "..", "..", "models.yml")
    if not os.path.exists(models_list_file):
        models_list_file = 'latest_silero_models.yml'
    if not os.path.exists(models_list_file):
        torch.hub.download_url_to_file('https://raw.githubusercontent.com/snakers4/silero-models/master/models.yml',
                                    'latest_silero_models.yml',
                                    progress=False)
    assert os.path.exists(models_list_file)
    models = OmegaConf.load(models_list_file)
    available_languages = list(models.stt_models.keys())
    assert language in available_languages

    model, decoder = init_jit_model(model_url=models.stt_models.get(language).get(version).get(jit_model),
                                    **kwargs)
    utils = (read_batch,
             split_into_batches,
             read_audio,
             prepare_model_input)

    return model, decoder, utils


def silero_tts(language='en',
               speaker='kseniya_16khz',
               **kwargs):
    """ Silero Text-To-Speech Models
    language (str): language of the model, now available are ['ru', 'en', 'de', 'es', 'fr']
    Returns a model and a set of utils
    Please see https://github.com/snakers4/silero-models for usage examples
    """
    from omegaconf import OmegaConf      
    from .tts_utils import apply_tts
    from .tts_utils import init_jit_model as init_jit_model_tts

    models_list_file = os.path.join(os.path.dirname(__file__), "..", "..", "models.yml")
    if not os.path.exists(models_list_file):
        models_list_file = 'latest_silero_models.yml'
    if not os.path.exists(models_list_file):
        torch.hub.download_url_to_file('https://raw.githubusercontent.com/snakers4/silero-models/master/models.yml',
                                    'latest_silero_models.yml',
                                    progress=False)
    assert os.path.exists(models_list_file)
    models = OmegaConf.load(models_list_file)
    available_languages = list(models.tts_models.keys())
    assert language in available_languages, f'Language not in the supported list {available_languages}'
    available_speakers = []
    speaker_language = {}
    for lang in available_languages:
        speakers = list(models.tts_models.get(lang).keys())
        available_speakers.extend(speakers)
        for _ in speakers:
            speaker_language[_] = lang
    assert speaker in available_speakers, f'Speaker not in the supported list {available_speakers}'
    assert language == speaker_language[speaker], f"Incorrect language '{language}' for this speaker, please specify '{speaker_language[speaker]}'"

    model_conf = models.tts_models[language][speaker].latest
    if '_v2' in speaker or '_v3' in speaker or 'v3_' in speaker or 'v4_' in speaker or 'v5_' in speaker:
        from torch import package
        model_url = model_conf.package
        model_dir = os.path.join(os.path.dirname(__file__), "model")
        os.makedirs(model_dir, exist_ok=True)
        model_path = os.path.join(model_dir, os.path.basename(model_url))
        if not os.path.isfile(model_path):
            torch.hub.download_url_to_file(model_url,
                                           model_path,
                                           progress=True)
        imp = package.PackageImporter(model_path)
        model = imp.load_pickle("tts_models", "model")
        if speaker == 'multi_v2':
            avail_speakers = model_conf.speakers
            return model, avail_speakers
        else:
            example_text = model_conf.example
            return model, example_text
    else:
        model = init_jit_model_tts(model_conf.jit)
        symbols = model_conf.tokenset
        example_text = model_conf.example
        sample_rate = model_conf.sample_rate
        return model, symbols, sample_rate, example_text, apply_tts


def silero_te():
    """ Silero Texts Enhancing Models
    Current model supports the following languages: ['en', 'de', 'ru', 'es']
    Returns a model and a set of utils
    Please see https://github.com/snakers4/silero-models for usage examples
    """
    import yaml
    from torch import package

    models_list_file = os.path.join(os.path.dirname(__file__), "..", "..", "models.yml")
    if not os.path.exists(models_list_file):
        models_list_file = 'latest_silero_models.yml'
    if not os.path.exists(models_list_file):
        torch.hub.download_url_to_file('https://raw.githubusercontent.com/snakers4/silero-models/master/models.yml',
                                    'latest_silero_models.yml',
                                    progress=False)

    assert os.path.exists(models_list_file)
    with open(models_list_file, 'r', encoding='utf8') as yaml_file:
        models = yaml.load(yaml_file, Loader=yaml.SafeLoader)
    model_conf = models.get('te_models').get('latest')

    model_url = model_conf.get('package')
    model_dir = os.path.join(os.path.dirname(__file__), "model")
    os.makedirs(model_dir, exist_ok=True)
    model_path = os.path.join(model_dir, os.path.basename(model_url))

    if not os.path.isfile(model_path):
        torch.hub.download_url_to_file(model_url,
                                       model_path,
                                       progress=True)

    imp = package.PackageImporter(model_path)
    model = imp.load_pickle("te_model", "model")
    example_texts = model.examples
    languages = model_conf.get('languages')
    punct = model_conf.get('punct')

    def apply_te(text, lan='en'):
        return model.enhance_text(text, lan)

    return (model,
            example_texts,
            languages,
            punct,
            apply_te)


def silero_denoise(
    name='small_slow',
    version='latest',
    **kwargs
):
    """ Silero Denoise Model(s)
    name (str): model name, available names (in order of best outputs to worst): ['small_slow', 'large_fast', 'small_fast']
    version (str): model version
    Returns a model and a set of utils
    Please see https://github.com/snakers4/silero-models for usage examples
    """
    from omegaconf import OmegaConf    
    from .denoiser_utils import (
        init_jit_model,
        read_audio,
        save_audio,
        denoise
    )

    models_list_file = os.path.join(os.path.dirname(__file__), "..", "..", "models.yml")
    if not os.path.exists(models_list_file):
        models_list_file = 'latest_silero_models.yml'
    if not os.path.exists(models_list_file):
        torch.hub.download_url_to_file(
            'https://raw.githubusercontent.com/snakers4/silero-models/master/models.yml',
            'latest_silero_models.yml',
            progress=False
        )
    assert os.path.exists(models_list_file)
    models = OmegaConf.load(models_list_file)
    available_models = models.denoise_models.models
    samples = models.denoise_models.samples
    assert name in available_models

    model = init_jit_model(model_url=models.denoise_models.get(name).get(version).get('jit'),
                                    **kwargs)
    utils = (
        read_audio,
        save_audio,
        denoise
    )

    return model, samples, utils
